from dataclasses import dataclass, field
from typing import Any, Callable


@dataclass
class FieldData:

    name: str
    value: Any
    annotation: Any


@dataclass
class Settings:
    """ Настройки для валидации.

        :validator:         Cсылка на функцию, которая будет использоваться
                            непосредственно для валидации.
                            Реализация поддерживаемых функций находится
                            в файле validators.py.
                            Можно сделать свою функцию для валидации и указать
                            ее в этом поле.
                            Функция "по умолчанию" установливается в модуле
                            декораторов при объявлении декоратора.

        :is_replace_args:   Если True, то будет производиться замена исходных
                            значений полей аргументов на экземпляры классов
                            валидации.

        :is_replace_result: Если True, то будет производиться замена значения
                            результата на экземпляры классов валидации.
        :extra:             Словарь с дополнительными значениями, который
                            будет передаваться в validator (например, можно
                            передать класс для валидации данных)
    """

    validator: Callable
    is_replace_args: bool = True
    is_replace_result: bool = True
    extra: dict = field(default_factory=dict)



""" Функции валидаторы."""

from dataclasses import fields, make_dataclass
from typing import Any, Dict, Optional

from pydantic import BaseModel, Extra, create_model, error_wrappers
from validated_dc import ValidatedDC


class ValidationError(Exception):
    pass


# Валидатор на основе pydantic.BaseModel -------------------------------------

# Префикс к именам полей, которые будут использоваться для создания
# валидирующего класса. Он необходим для предотвращения конфликта имен.
NAME_PREFIX = "field_name_pref_"


class ModelForValidation(BaseModel):
    """ Класс для валидации по умолчанию."""

    class Config:
        # Строгая проверка имен параметров при создании экземпляра
        extra = Extra.forbid
        # Разрешение для пользовательских типов
        arbitrary_types_allowed = True


def pydantic_validator(
    annotations: Dict[str, Any], values: Dict[str, Any],
    is_replace: bool, extra: dict
) -> Optional[Dict[str, Any]]:
    """ Функция для проверки соответствия значений полей их аннотациям.

        :annotations: Словарь, который содержит имена полей и их аннотации.
        :values:      Словарь, который содержит имена полей и их значения.

        :is_replace:  Если True, то функция вернет словарь с именами
                      отвалидированных полей и их значениями после валидации.
                      Таким образом, если у поля была аннотация с наследником
                      (наследниками) `BaseModel` и они поступили на валидацию
                      в виде словаря (в виде словарей, например если пришел
                      список словарей), то они будут заменены на экземпляр
                      (на экземпляры) `BaseModel`, и в вызывающем коде к
                      ним можно будет обращаться "через точку".
                      Если параметр равен False, то функция вернет None.

        :extra:       Словарь с дополнительными параметрами.
                      Если в параметре `extra` имеется ключ `base_val_class`
                      то его значение должно быть наследником BaseModel,
                      и этот класс будет использоваться для валидации.
                      Если в словаре `extra` НЕТ ключа `base_val_class`,
                      то для валидации используется класс ModelForValidation.
    """

    base_val_class = extra.get("base_val_class")
    if base_val_class is None:
        base_val_class = ModelForValidation

    kwargs = {"__base__": base_val_class}

    for field_name, field_annotation in annotations.items():
        kwargs[NAME_PREFIX+field_name] = (field_annotation, ...)

    ValidatorClass = create_model("argument with the name of:", **kwargs)

    try:
        instance = ValidatorClass(
            **{NAME_PREFIX+k: v for k, v in values.items()}
        )

    except error_wrappers.ValidationError as error:
        error_str = str(error).replace(NAME_PREFIX, "")
        raise ValidationError(error_str)

    return {
        name.replace(NAME_PREFIX, ""): value
        for name, value in dict(instance).items()
        # TODO Сделать фильтр для полей содержащих экземпляры BaseModel
    } if is_replace else None


# Валидатор на основе validated_dc.ValidatedDC -------------------------------

def validated_dc_validator(
    annotations: Dict[str, Any], values: Dict[str, Any],
    is_replace: bool, extra: dict
) -> Optional[Dict[str, Any]]:
    """ Функция для проверки соответствия значений полей их аннотациям.

        :annotations: Словарь, который содержит имена полей и их аннотации.
        :values:      Словарь, который содержит имена полей и их значения.

        :is_replace:  Если True, то функция вернет словарь с именами
                      отвалидированных полей и их значениями после валидации.
                      Таким образом, если у поля была аннотация с наследником
                      (наследниками) `ValidatedDC` и они поступили на валидацию
                      в виде словаря (в виде словарей, например если пришел
                      список словарей), то они будут заменены на экземпляр
                      (на экземпляры) `ValidatedDC`, и в вызывающем коде к
                      ним можно будет обращаться "через точку".
                      Если параметр равен False, то функция вернет None.

        :extra:       Словарь с дополнительными параметрами.
                      Если в параметре `extra` имеется ключ `base_val_class`
                      то его значение должно быть наследником BaseModel,
                      и этот класс будет использоваться для валидации.
                      Если в словаре `extra` НЕТ ключа `base_val_class`,
                      то для валидации используется класс ValidatedDC.
    """

    base_val_class = extra.get("base_val_class")
    if base_val_class is None:
        base_val_class = ValidatedDC

    ValidatorClass = make_dataclass(
        "ValidatorClass", list(annotations.items()),
        bases=(base_val_class, )
    )

    instance: ValidatedDC = ValidatorClass(**values)

    errors = instance.get_errors()
    if errors is not None:
        raise ValidationError(str(errors))

    return {
        field.name: getattr(instance, field.name) for field in fields(instance)
        # TODO Сделать фильтр для полей содержащих экземпляры ValidatedDC
    } if is_replace else None



import inspect
import logging
from typing import Any, Callable, Dict, List, Tuple

from valdec.data_classes import FieldData, Settings

logger = logging.getLogger()


def get_data_with_annotations(
    func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any]
) -> List[FieldData]:
    """ Связывает аргументы функции имеющие аннотации с полученными(!) данными.

        Возвращает список с полной информацией о каждом полученном значении,
        а именно: имя аргумента, полученное значение, аннотация из сигнатуры
        функции (см. FieldData).
    """

    signature = inspect.signature(func)
    parameters = signature.parameters
    bound = signature.bind(*args, **kwargs)

    return [
        FieldData(name, value, parameters[name].annotation)
        for name, value in bound.arguments.items()
        if parameters[name].annotation is not inspect._empty
    ]


def get_data_for_validation(
    fields: List[FieldData], names: Tuple[Any, ...], exclude: bool
) -> List[FieldData]:
    """ Формирует и возвращает окончательный список данных для валидации.

        :fields:  Список с полной информацией о каждом поле (имя из сигнатуры
                  функции, полученное значение, аннотация).
        :names:   Имена полей подлежащих включению или исключению из валидации.
        :exclude: Если `True`, то для валидации будут отобраны поля из
                  `fields` с именами которых нет в `names`.
                  Если `False`, то будут отобраны поля из `fields` имена
                  которых есть в `names`. Но если при этом `names` пустой, то
                  будут отобраны все поля из fields.
    """

    if exclude:
        return [field for field in fields if field.name not in names]
    else:
        if names:
            return [field for field in fields if field.name in names]
        else:
            return fields


def get_annotations_values_dicts(
    fields: List[FieldData],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """ Возвращает словари с аннотациями и значениями полей."""

    annotations = {}
    values = {}

    for field in fields:
        field_name = field.name
        annotations[field_name] = field.annotation
        values[field_name] = field.value

    return annotations, values


class ValidationError(Exception):
    pass


class ValidationArgumentsError(ValidationError):
    pass


class ValidationReturnError(ValidationError):
    pass


def run_validation(
    fields: List[FieldData],
    validator: Callable,
    is_replace: bool,
    extra: dict,
    is_arguments: bool,
) -> Dict[str, Any]:
    """ Преобразует список полей в словари, запускает валидацию и возвращает
        ее результат.
    """
    try:
        annotations, values = get_annotations_values_dicts(fields)
        result = validator(
            annotations, values, is_replace, extra
        )
    except Exception as error:

        error_class = ValidationArgumentsError if is_arguments \
            else ValidationReturnError

        raise error_class(f"Validation error {type(error)}: {str(error)}.")

    return result


def replace_args_kwargs(
    func: Callable, args: tuple, kwargs: Dict[str, Any],
    replaceable_arguments: Dict[str, Any],
) -> Tuple[tuple, Dict[str, Any]]:
    """ Производит замену значений аргументов из args и kwargs на значения
        из replaceable_arguments (только тех, которые есть в
        replaceable_arguments) и возвращает новые args и kwargs для функции.

        :func:   Сылка на функцию, для сигнатуры которой будут создаваться
                 новые значения для позиционных и именованных аргументов.
        :args:   Кортеж с исходными значениями позиционных аргументов
                 предназначавшихся для передачи в функцию.
        :kwargs: Словарь с исходными значениями именованных аргументов
                 предназначавшихся для передачи в функцию.

        :replaceable_arguments: Словарь с именами аргументов и их новыми
                                значениями.
    """

    parameters = inspect.signature(func).parameters

    parameters_keys = list(parameters.keys())
    new_args = list(args)

    for key in replaceable_arguments.keys():

        if kwargs.get(key) is None:
            pos = parameters_keys.index(key)
            new_args[pos] = replaceable_arguments[key]
        else:
            kwargs[key] = replaceable_arguments[key]

    return tuple(new_args), kwargs


def get_names_from_decorator(names_or_func: tuple) -> tuple:
    """ Если в полученном кортеже первый параметр функция или кортеж пустой,
        то вернет пустой кортеж.
        Иначе вернет исходный кортеж.

        Функция нужна для корректой работы декоратора при различных способах
        его вызова.
    """

    names_from_decorator = names_or_func
    if not names_or_func or callable(names_or_func[0]):
        names_from_decorator = tuple()

    return names_from_decorator


def before(
    func: Callable, args: tuple, kwargs: Dict[str, Any],
    names_or_func: Any, exclude: bool, settings: Settings,
) -> Tuple[tuple, Dict[str, Any]]:
    """ Часть декоратора которая валидирует входящие аргументы декорируемой
        функции.
        Выполняется до вызова декорируемой функции.

        Получает функцию, все её аргументы, и все аргументы декоратора.

        Возвращает аргументы для вызова функции (args и kwargs, возможно
        измененные).
    """

    names_from_decorator = get_names_from_decorator(names_or_func)

    data_with_annotation = get_data_with_annotations(
        func, args, kwargs
    )
    if data_with_annotation:
        data_for_validation = get_data_for_validation(
            data_with_annotation, names_from_decorator, exclude
        )
        if data_for_validation:

            logger.info(f"Going to validate arguments: {data_for_validation}")

            replaceable_args = run_validation(
                data_for_validation,
                settings.validator,
                settings.is_replace_args,
                settings.extra,
                is_arguments=True,
            )
            if replaceable_args is not None:

                logger.info(f"Replace: {replaceable_args}")

                args, kwargs = replace_args_kwargs(
                    func, args, kwargs, replaceable_args
                )
    return args, kwargs


def after(
    func: Callable, result: Any,
    names_or_func: Any, exclude: bool, settings: Settings
) -> Any:
    """ Часть декоратора которая валидирует результат декорируемой функции.
        Выполняется после вызова декорируемой функции.

        Получает функцию, её результат, и все аргументы декоратора.

        Возвращает результат функции (возможно измененный).
    """

    names_from_decorator = get_names_from_decorator(names_or_func)

    annotation = func.__annotations__.get("return")
    if annotation is None:
        annotation = type(None)

    data_with_annotation = [FieldData("return", result, annotation), ]

    data_for_validation = get_data_for_validation(
        data_with_annotation, names_from_decorator, exclude
    )
    if data_for_validation:

        data_for_validation = [FieldData("result", result, annotation), ]

        logger.info(f"Going to validate: {data_for_validation}")

        replaceable = run_validation(
            data_for_validation,
            settings.validator,
            settings.is_replace_result,
            settings.extra,
            is_arguments=False,
        )
        if replaceable is not None:

            logger.info(f"Replace: {replaceable}")

            result = replaceable["result"]

    return result
